/* $Id: HFWorld.cpp,v 1.8 2002/12/27 20:16:46 mat Exp $
**
** Ark - Libraries, Tools & Programs for MMORPG developpements.
** Copyright (C) 1999-2000 The Contributors of the Ark Project
** Please see the file "AUTHORS" for a list of contributors
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
 #include <windows.h>
#endif

#include <GL/gl.h>

#include <algorithm>
#include <iostream>
#include <sstream>

// For the random stuff 
#include <stdlib.h>
// For memcpy 
#include <string.h>

#include <Ark/ArkConfig.h>
#include <Ark/ArkRenderer.h>
#include <Modules/HeightField/HFWorld.h>
#include <Modules/HeightField/HFQuadtree.h>
#include <Modules/HeightField/SkyDome.h>

namespace Ark
{

   // Allocate a new "size*size" height field
   HeightField::HeightField (Cache *cache, WorldUpdater *upd) : 
       World (cache,upd),
       m_OffsetX (0.0f), 
       m_OffsetZ (0.0f),
       m_SizeX (0), 
       m_SizeZ (0),
       m_Scale (0.25f),
       m_ScaleY(0.0125f),
       m_Y (0),
       m_G (0),
      // By default there's no pathfinder (it takes a lot of memory)
      // call InitPathfinder() to enable it...
       m_Pathfinder (0),
      // No quadtree either :)
       m_Quadtree (0),
       m_Skybox (0)
   {
   }


   // Destroy the given Height-field.
   HeightField::~HeightField ()
   {
      delete[] m_Y;
      delete[] m_G;

      DestroyPathfinder();

      delete m_Quadtree;
      delete m_Skybox;

      MaterialList::iterator it;
      for (it = m_Grounds.begin(); it != m_Grounds.end(); ++it)
	 (*it)->Unref();
   }

   /**
    * Some notes on the game data : in the directory pointed to by \c path,
    * there should be a configuration file called world.cfg, which contains
    * general informations about the world, and informations about the
    * heightfield.
    * The variable \c heightfield::HeightData should point to a 8bit gray
    * texture containing elevation data. There should also be a
    * heightfield::GroundData pointing to a 8bit gray image, each level of
    * gray being an index in the groundtype list. The last thing needed is
    * a list of ground types present in the map.
    */
   bool
   HeightField::Load (const String &path)
   {
      String cfg = path + "/world.cfg";

      /// XXX not implemented: m_Config.Clear();

      if (! m_Config.Load( cfg ) )
	  return false;

      m_Skybox = new SkyBox(m_Cache, m_Config);

      m_OffsetX = m_Config.GetScalar ("heightfield::OffsetX", 0.0f);
      m_OffsetZ = m_Config.GetScalar ("heightfield::OffsetZ", 0.0f);

      m_Scale = m_Config.GetScalar ("heightfield::Scale", 1.0f);
      m_ScaleY = m_Config.GetScalar ("heightfield::ScaleY", 1.0f);

      /// Load elevation map.
      String hdata = m_Config.GetStr("heightfield::HeightData", String());

      Image *elevation;
      m_Cache->Get (V_IMAGE, hdata, &elevation);

      if (elevation == 0 || elevation->m_Format != Image::GRAY_8)
	 return false;

      m_SizeX = elevation->m_Width;
      m_SizeZ = elevation->m_Height;

      // used for several for loops (MSVC caveat)
      unsigned int i;

      // constant size of loaded images
      const unsigned int size = m_SizeX * m_SizeZ;
      
      // allocate memory for height field
      m_Y = new Height[size];
      
      for (i = 0; i < size; i++)
		  m_Y[i] = elevation->m_Data[i];

      elevation->Unref();

      /// Load ground map.
      Image *groundmap;
      m_Cache->Get (V_IMAGE,
		    m_Config.GetStr("heightfield::GroundData",""),
		    &groundmap);

      if (groundmap == 0 || groundmap->m_Format != Image::GRAY_8)
		  return false;

      if (groundmap->m_Width != m_SizeX ||
	  groundmap->m_Height != m_SizeZ)
		  return false;


      // Copy grounds
      m_G = new uchar[size];
      memcpy (m_G, groundmap->m_Data, size);
      groundmap->Unref();
      
      const size_t ngrounds = m_Config.GetInt ("heightfield::NumGrounds", 0);

      for (i = 0; i < size; i++)
      {
	  if (ngrounds <= m_G[i])
	      m_G[i] = 0;
      }

      /// Load ground list.

      m_Grounds.resize (ngrounds);
      for (size_t j = 0; j < ngrounds; j++)
      {
	  std::ostringstream os;
	  os << "heightfield::Ground" << j;
	  String grd = os.str();

	  m_Cache->Get(V_MATERIAL, m_Config.GetStr(grd, String()), &m_Grounds[j]);
      }


      return true;
   }

   // ======
   // Write the hf to a file
   // ======
   bool
   HeightField::Write (const String &path, bool mipmaps)
   {
      assert (!"todo Write");
      return false;
   }

   bool HeightField::Init (int init_flags)
   {
     if (((init_flags & WORLD_HAS_RENDERING) ||
           (init_flags & WORLD_HAS_COLLISION)) &&
         m_Quadtree == 0)
     {
       m_Quadtree = new Quadtree(this);
     }


     if ((init_flags & WORLD_HAS_PATHFINDING)
         && m_Pathfinder == 0)
     {
       InitPathfinder ();
     }

     return true;
   }

// ======
// Returns the height (y-value) of a point in this heightmap. The given (x,z)
// are in world coordinates. Heights outside this heightmap are considered to
// be 0. Heights between sample points are bilinearly interpolated from
// surrounding points.
// ======
   scalar
   HeightField::GetHeight (scalar xp_, scalar zp_) const
   {
       // Goes into HF coords
       const scalar xp = xp_ - m_OffsetX;
       const scalar zp = zp_ - m_OffsetZ;

       // Gets rid of negative values (return 0)
       if ( (xp < 0.0f) || (zp < 0.0f) )
	   return 0.0f;

      const scalar ooscl = 1.0f / m_Scale;
      const scalar x = xp * ooscl;
      const scalar z = zp * ooscl;

      const unsigned int fx = static_cast< unsigned int >( x );
      const unsigned int fz = static_cast< unsigned int >( z );

      // Inside the heightfield
      if ( (fx < m_SizeX) && (fz < m_SizeZ) )
      {
	  const scalar lx = x - fx;
	  const scalar lz = z - fz;

	  // Bilinearly interpolate from surrounding points
	  const scalar x0 =  Y(fx,   fz) + ( Y(fx,   fz+1) - Y(fx,   fz) ) * lz;
	  const scalar x1 =  Y(fx+1, fz) + ( Y(fx+1, fz+1) - Y(fx+1, fz) ) * lz;
	  
	  return ( x0 + (x1 - x0)*lx ) * m_ScaleY;
      }
      
      return 0.0f;
   }

   Vector3 
   HeightField::GetRestPosition(const Vector3& position) const
   {
       Vector3 result(position);
       result.Y = GetHeight(position.X, position.Z);
       return result;
   }
   
   void
   HeightField::DeleteGround (unsigned int cur)
   {
     const unsigned int nitems = m_Grounds.size();
     const unsigned int sz = m_SizeX * m_SizeZ;
     uchar *grd = &GetGround(0,0);

     m_Grounds[cur]->Unref();
     m_Grounds.erase(m_Grounds.begin()+cur);

     if (cur == nitems-1)
     {
       for (unsigned int i = 0; i < sz; i++)
       {
         if (grd[i] == cur)
           grd[i] = nitems-1;
       }
     }
     else if (cur > 0)
     {
       for (unsigned int i = 0; i < sz; i++)
       {
         if (grd[i] > cur)
           grd[i]--;
       }
     }
     else
     {
       for (unsigned int i = 0; i < sz; i++)
       {
         if (grd[i])
           grd[i]--;
       }
     }

     if (m_Quadtree)
       m_Quadtree->SetMaterials (m_Grounds);
   }

   void
   HeightField::AppendGround (Material *mat)
   {
      m_Grounds.push_back(mat);
      mat->Ref();

      if (m_Quadtree)
        m_Quadtree->SetMaterials(m_Grounds);
   }
   
   /**
    * Calls invalidate on the quadtree
    */
   void
   HeightField::Invalidate (scalar minx, scalar minz, scalar maxx, scalar maxz)
   {
       if (m_Quadtree)
	   m_Quadtree->Invalidate(minx, minz, maxx, maxz);
   }

   // =======================================================================
   
   bool g_DrawTerrain = true;
   bool g_DrawEntities = true;
   bool g_DrawSkels = false;

   // =======================================================================

   /// FIXME FIXME FIXME FIXME FIXME
   /// THE FOLLOWING CODE IS A PIECE OF JUNK

   /// UGHH !!! Quite awful, isn't it ?? FIXME: Lights don't have to be
   /// hardcoded  that way, that was just for testing purposes.
   static void SetLight(void)
   {
      GLfloat ambient[] = {0.4f, 0.4f, 0.4f, 1.0f};
      glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

      GLfloat light0Pos[4] = {128.0f, 5.0f, 128.0f, 0.0f};
      GLfloat light0Amb[4] = {0.3f, 0.3f, 0.3f, 1.00f};
      GLfloat light0Diff[4] = {1.00f, 1.00f, 1.00f, 1.00f};
      GLfloat light0Spec[4] = {1.00f, 1.00f, 1.00f, 1.00f};
      GLfloat light0SpotDir[3] = {0.00f, -1.00f, 0.00f};
      
      GLfloat light0SpotExp = 0.00f;
      GLfloat light0SpotCutoff = 180.00f;
      GLfloat light0Atten0 = 1.00f;
      GLfloat light0Atten1 = 0.00f;
      GLfloat light0Atten2 = 0.00f;
      
      glEnable(GL_LIGHT0);
      glLightfv(GL_LIGHT0, GL_POSITION, light0Pos);
      glLightfv(GL_LIGHT0, GL_AMBIENT, light0Amb);    
      glLightfv(GL_LIGHT0, GL_DIFFUSE, light0Diff);
      glLightfv(GL_LIGHT0, GL_SPECULAR, light0Spec);

      glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, light0SpotDir);
      glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, light0SpotExp);
      glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, light0SpotCutoff);

      glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, light0Atten0);
      glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, light0Atten1);
      glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, light0Atten2);
   }

   void SetFog()
   {
#if 0
      glEnable(GL_FOG);

      GLfloat fogColor[4] = {0.8, 0.4, 0.5, 1.0};

      GLenum fogMode = GL_EXP;
      glFogi (GL_FOG_MODE, fogMode);
      glFogfv (GL_FOG_COLOR, fogColor);
      glFogf (GL_FOG_DENSITY, 0.01);
      glHint (GL_FOG_HINT, GL_DONT_CARE);
      glFogf (GL_FOG_START, 10.0);
      glFogf (GL_FOG_END, 40.0);
#endif
   }

   void UnsetFog()
   {
#if 0
      glDisable(GL_FOG);
#endif
   }

   void
   DrawModelSkel (Skeleton* mdl, Ark::Matrix44 *matrices)
   {
      bool enab = glIsEnabled(GL_TEXTURE_2D);
      bool enabdt = glIsEnabled(GL_DEPTH_TEST);
      if (enab) glDisable (GL_TEXTURE_2D);
      if (enabdt) glDisable (GL_DEPTH_TEST);

      for (size_t i = 0; i < mdl->m_Bones.size(); i++)
      {
         Bone &bone = mdl->m_Bones[i];
	 
         if (bone.m_Parent >= 0)
         {
            glPointSize (3.0f);
            glColor3f (1, 0.7f, 0);

            glBegin (GL_LINES);
            glVertex3f (matrices[bone.m_Parent].M(3,0),
                        matrices[bone.m_Parent].M(3,1),
                        matrices[bone.m_Parent].M(3,2));
            glVertex3f (matrices[i].M(3,0),
                    matrices[i].M(3,1),
                        matrices[i].M(3,2));
            glEnd ();

            // Draw points
            glColor3f (0, 0, 0.8f);
            glBegin (GL_POINTS);
            if (mdl->m_Bones[bone.m_Parent].m_Parent != -1)
               glVertex3f (matrices[bone.m_Parent].M(3,0),
                           matrices[bone.m_Parent].M(3,1),
                           matrices[bone.m_Parent].M(3,2));
            glVertex3f (matrices[i].M(3,0),
                        matrices[i].M(3,1),
                        matrices[i].M(3,2));
            glEnd ();
         }
         else
         {
            // Draw parent bone node
            glPointSize (5.0f);
            glColor3f (0.8f, 0, 0);

            glBegin (GL_POINTS);
            glVertex3f (matrices[i].M(0,3),
                        matrices[i].M(1,3),
                        matrices[i].M(2,3));
            glEnd ();
         }
      }

      glPointSize (1.0f);
      if (enab) glEnable(GL_TEXTURE_2D);
      if (enabdt) glEnable (GL_DEPTH_TEST);
   }

   bool
   HeightField::Render (Renderer &renderer, const Camera& camera)
   {
      if (m_Quadtree == NULL)
	 return false;

      scalar height = GetHeight (camera.m_PointOfView.X,
				 camera.m_PointOfView.Z);

      Camera copyCamera = camera;

      copyCamera.m_PointOfView.Y = std::max(camera.m_PointOfView.Y, height + 5.0f);

      renderer.SetCamera( copyCamera );

      if (m_Skybox)
	 m_Skybox->Render(renderer, camera.m_PointOfView);


      SetLight();
      SetFog();

      // === Render the world
      if (g_DrawTerrain) 
      {
	  m_Quadtree->Render(renderer, camera);
      }

      const std::vector<Entity*>& entities = GetEntities();
      std::vector<Entity*>::const_iterator cur;

      if (g_DrawEntities)
      {	 
	  // === Render the entities' models ===
	  const Frustum &frustum = renderer.GetFrustum();

	  for (cur=entities.begin() ; cur != entities.end() ; ++cur)
	  {
	      if ((*cur)->m_MState.GetModel())
	      {
		  // Cull invisible entities using view frustum
		  if (frustum.GetVisibility ((*cur)->m_BBox) != OUTSIDE)
		      (*cur)->m_MState.Render(renderer);

		  if (g_DrawSkels &&(*cur)->m_MState.GetModel()->m_Skeleton)
		     DrawModelSkel((*cur)->m_MState.GetModel()->m_Skeleton,
				   (*cur)->m_MState.m_BoneMatrices);
	      }
	  }
      }

      // === Draw the particle systems  ===
      // I do this without z buffer writing so a particle won't occlude another
      // at a point where it is transparent. It also avoid having to sort
      // the particle in Z order, which can be quite time consuming.
      // Finally, it doesn't matter visually...
      
      // === Draw the "athmospheric" particle system (snow, rain, etc) ===
      // if (m_System != NULL)
      // {
      //  m_System->Update (dt);
      //  m_System->Render ();
      // }
      
      // === Draw the particle systems bound to the entity ===
      for (cur=entities.begin() ; cur != entities.end() ; ++cur)
      {
	  std::vector<EntityPSys*> list = (*cur)->m_Particles;
	  std::vector<EntityPSys*>::iterator i;

	  for (i=list.begin() ; i != list.end(); ++i)
	  {
	      if (*i == NULL)
		  continue;

	      (*i)->Render (renderer);
	  }

      }
   
      // === Draw the entity path (debugging) ===
      for (cur=entities.begin(); cur != entities.end(); ++cur)
      {
	 const std::list<Vector3> path = (*cur)->m_Path.GetPoints();

	 if (path.empty())
	    continue;

	 glColor3f (1.0, 0.0, 0.0);
	 glLineWidth (3.0);
	 glBegin (GL_LINES);
	 
	 glVertex3fv (&path.front().X);
	    
	 std::list<Vector3>::const_iterator i;
	 for (i = path.begin(); i != path.end(); ++i)
	 {
	    glVertex3fv (&i->X);
	    glVertex3fv (&i->X);
	 }
	 
	 glVertex3fv (&path.back().X);
	 
	 glEnd ();
	 glLineWidth (1.0);
      }
      glColor4f (1.0, 1.0, 1.0, 1.0);

      UnsetFog();

      return true;
   }

/* namespace Ark */
}

#define ARK_MODULE
#include <Ark/ArkFactoryReg.h>
class HFWorldFactory : public Ark::WorldFactory
{
   public:
      virtual ~HFWorldFactory(){}
      virtual Ark::World *NewWorld(Ark::Cache *cache,
				   Ark::WorldUpdater *updater)
      { return new Ark::HeightField(cache, updater); }
};

ARK_REGISTER("ark::World::HeightField", HFWorldFactory);
